from pydantic import BaseModel, field_validator



class MsgLevel:
    INFO = "info"
    SUCCESS =  "success"
    WARNING = "warning"
    ERROR =  "error"
    ALL = [INFO, SUCCESS, WARNING, ERROR]
    
    @classmethod
    def is_valid_level(cls, level: str) -> bool:
        return level in cls.ALL
    
    @classmethod
    def get_levels(cls) -> list[str]:
        return cls.ALL
    
    @classmethod
    def get_level_names(cls) -> list[str]:
        return [level.upper() for level in cls.ALL]
    
    @classmethod
    def get_level_map(cls) -> dict[str, str]:
        return {level: level.upper() for level in cls.ALL}
    
    @classmethod
    def raise_if_invalid(cls, level: str) -> None:
        if not cls.is_valid_level(level):
            raise ValueError('level must be one of: ' + ', '.join(cls.ALL))
        
        
class MsgModel(BaseModel):
    level: str = MsgLevel.ERROR
    message: str

    @field_validator('level')
    def validate_level(cls, v):
        if v not in MsgLevel.ALL:
            raise ValueError('level must be one of: ' + ', '.join(MsgLevel.ALL))
        return v
    
    def is_info(self) -> bool:
        return self.level == MsgLevel.INFO
    
    def is_success(self) -> bool:
        return self.level == MsgLevel.SUCCESS
    
    def is_warning(self) -> bool:
        return self.level == MsgLevel.WARNING
    
    def is_error(self) -> bool:
        return self.level == MsgLevel.ERROR
    
    def __str__(self) -> str:
        return f"[{self.level.upper()}] [{self.level}] {self.message}"
    
    def __repr__(self) -> str:
        return f"MsgModel(level={self.level}, level={self.level}, message={self.message})"
    
    
class MsgFactory:
    
    @classmethod
    def create(cls, level: str, message: str) -> MsgModel:
        MsgLevel.raise_if_invalid(level)
        return MsgModel(level=level, message=message)


from typing import Self
from nkunyim_util.models.msg_model import MsgModel


class MsgBag:
    
    def __init__(self) -> None:
        self.msgs = []
    
    def add(self, error: MsgModel) -> None:
        self.msgs.append(error)
        
    def get(self) -> list[MsgModel]:
        return self.msgs
    
    def has_msgs(self) -> bool:
        return len(self.msgs) > 0
    
    def clear(self) -> None:
        self.msgs = []
        
    def extend(self, other: Self) -> None:
        self.msgs.extend(other.msgs)
        
    def merge(self, other: Self) -> None:
        self.extend(other)
        
    def count(self) -> int:
        return len(self.msgs)
        
    def copy(self):
        new_service = MsgBag()
        new_service.msgs = self.msgs.copy()
        return new_service
    
    def reset(self) -> None:
        self.clear()
        
    def is_empty(self) -> bool:
        return not self.has_msgs()
    
    def first(self) -> MsgModel | None:
        if self.has_msgs():
            return self.msgs[0]
        return None
    
    def last(self) -> MsgModel | None:
        if self.has_msgs():
            return self.msgs[-1]
        return None
    
    def messages(self) -> list[str]:
        return [error.message for error in self.msgs]
        
    def levels(self) -> list[str]:
        return [error.level for error in self.msgs]
    
    def to_dict_list(self) -> list[dict]:
        return [error.model_dump() for error in self.msgs]
    
    def from_dict_list(self, dict_list: list[dict]) -> None:
        self.msgs = [MsgModel(**data) for data in dict_list]
        
    def __len__(self) -> int:
        return len(self.msgs)
    
    def __iter__(self):
        return iter(self.msgs)
    
    def __getitem__(self, index: int) -> MsgModel:
        return self.msgs[index]
    
    def __bool__(self) -> bool:
        return self.has_msgs()    
    
    def __str__(self) -> str:
        return ", ".join([f"[{error.level}] {error.message}" for error in self.msgs])
    
    def __repr__(self) -> str:
        return f"MsgBag(msgs={self.msgs})"


from decimal import Decimal
from uuid import UUID
from pydantic import BaseModel

 
class ContinentModel(BaseModel):
    id: UUID
    code: str
    name: str


class CurrencyModel(BaseModel):
    id: UUID
    code: str
    name: str
    symbol: str
    decimals: int


class NationModel(BaseModel):
    id: UUID
    code: str
    name: str
    phone: str
    continent: ContinentModel
    currency: CurrencyModel
    capital: str
    languages: str
    north: Decimal
    south: Decimal
    east: Decimal
    west: Decimal
    flag: str
    flag_2x: str
    flag_3x: str
    flag_svg: str
    is_active: bool



from datetime import datetime
from pydantic import BaseModel


class TokenModel(BaseModel):
    access_token: str
    refresh_token: str
    token_type: str
    expires_in: int
    expires_at: datetime
    scope: str



from nkunyim_util.bags.msg_bag import MsgBag


class ViewBag:

    def __init__(self):
        self.ok = False
        self.msgs = MsgBag()
        self.data = {}
        
    def set_ok(self, ok: bool) -> None:
        self.ok = ok
        
    def is_ok(self) -> bool:
        return self.ok
    
    def add_msg(self, msg) -> None:
        self.msgs.add(msg)
        
    def get_msgs(self) -> MsgBag:
        return self.msgs
    
    def set_data(self, key: str, value) -> None:
        self.data[key] = value
        
    def get_data(self, key: str):
        return self.data.get(key, None)
    
    def get_all_data(self) -> dict:
        return self.data
    
    def clear_data(self) -> None:
        self.data = {}
        
    def to_dict(self) -> dict:
        return {
            "ok": self.ok,
            "msgs": self.msgs.to_dict_list(),
            "data": self.data
        }


import json
import os
from typing import Union

from django.conf import settings
from nkunyim_util.management.commands.commands import DATAX_MIGRATE_EXCEPTION, DATAX_REVERSE_EXCEPTION
from nkunyim_util.management.commands.migrations import BaseDataxMigration



class DataxMigration(BaseDataxMigration):
    
    def migrate(self) -> Union[str, None]:
        super().migrate()
        
        try:
            # Write your data migration.migrate functions here.
              
            data = {
                "id": "29eb115f-aab8-406b-b6f6-f9050c0bec56",
                "code": "GH",
                "name": "Ghana",
                "phone": "233",
                "capital": "Accra",
                "languages": "en-GH,ak,ee,tw",
                "north": 11.173301,
                "south": 4.736723,
                "east": 1.191781,
                "west": -3.25542,
                "is_active": True,
            }
            
            base_path = os.path.join(settings.BASE_DIR, 'nkunyim_data/data')
            file_path = os.path.join(base_path, "nations.json")
                
            # Create Datax root directory if not exist.
            os.makedirs(base_path, exist_ok=True)
        
            with open(file_path, "w") as json_file:
                json.dump(data, json_file, indent=4, sort_keys=True)
                
            # Return None if successfull
            return None
        except IOError as ex:
            return f"{DATAX_MIGRATE_EXCEPTION}: {ex}"
        except Exception as ex:
            return f"{DATAX_MIGRATE_EXCEPTION}: {ex}"
        
    
    def reverse(self) -> Union[str, None]:
        super().reverse()
        
        try:
            # Write your data migration.reverse functions here
        
        
            # Return None if successfull
            return None
        except Exception as ex:
            return f"{DATAX_REVERSE_EXCEPTION}: {ex}"
    
    



from typing import Optional
from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.bags.msg_bag import MsgBag
from nkunyim_util.models.user_model import UserModel
from nkunyim_util.api.nkunyim_api_client import NkunyimApiClient
from nkunyim_util.services.session_service import SessionService
from nkunyim_util.services.signals import user_data_updated

class UserService:
    
    def __init__(self, req: HttpRequest) -> None:
        self.req = req
        self.msgs = MsgBag()
        
        
    def api(self, token: str) -> Optional[UserModel]:
        ...
    
    def app(self) -> Optional[UserModel]:      
        try:
            session_service = SessionService(req=self.req)
            if not session_service.has_lifetime:
                return None
            
            user_model = session_service.get_user()
            if user_model:
                return user_model
            
            client = NkunyimApiClient(req=self.req, name=settings.IDENTITY_SERVICE_NAME)
            response = client.get(path="/api/users/me")
            user_data = response.json() if response.ok else None  
            if not user_data:
                return None
            
            if not ('id' in user_data and user_data['id']):
                return None
            
            user_model = UserModel(**user_data)
            session_service.set_user(user_model=user_model)
            
            # Inform interested parties
            user_data_updated.send(sender=UserModel, instance=user_model)
            
            return user_model
        except:
            return None 
        



from datetime import datetime
from typing import List, Optional
from django.http import HttpRequest

from nkunyim_util.models.session_model import SessionModel, TokenModel
from nkunyim_util.models.user_model import UserModel


class SessionService:
    
    def __init__(self, req: HttpRequest) -> None:
        self.req = req
    
    
    def _key(self) -> str:
        parts = self.req.get_host().lower().split('.')
        if len(parts) >= 2:
            domain = '.'.join(parts[-2:])
            subdomain = parts[-3] if len(parts) > 2 else "www"
            return f"{subdomain}.{domain}"
        return "www.localhost"
    
    
    def set(self, model: SessionModel) -> None:
        self.req.session[self._key()] = model.model_dump()
        self.req.session.modified = True
        
        
    def get(self) -> Optional[SessionModel]:
        key = self._key()
        if not bool(key in self.req.session):
            return None

        session = self.req.session[key]
        return SessionModel(**session)
    
    
    def clear(self) -> None:
        self.req.session.clear()
        self.req.session.flush()
    
    
    def get_user(self) -> Optional[UserModel]:
        session_model = self.get()
        if session_model:
            return session_model.user
        
        
    def set_user(self, user_model: UserModel) -> None:
        session_model = self.get()
        if session_model:
            session_model.user = user_model
        else:
            session_model = SessionModel(user=user_model, token=None)
            
        self.set(model=session_model)
            
        
    def get_token(self) -> Optional[TokenModel]:
        session_model = self.get()
        if session_model:
            return session_model.token
            
        
    def set_token(self, token_model: TokenModel) -> None:
        session_model = self.get()
        if session_model:
            session_model.token = token_model
        else:
            session_model = SessionModel(user=None, token=token_model)
            
        self.set(model=session_model)
    
    
    def has_perms(self, perms: List[str]) -> bool:
        user_model = self.get_user()
        if not (user_model and user_model.perms):
            return False
    
        for perm in perms:
            if perm not in user_model.perms:
                return False
        return True
            
    
    @property
    def lifetime(self) -> Optional[int]:
        token = self.get_token()
        if not token:
            return None
        current_datetime = datetime.now()
        time_difference = token.expires_at - current_datetime
        difference_in_seconds = time_difference.total_seconds()
        return int(difference_in_seconds)


    @property
    def has_lifetime(self) -> bool:
        return bool(self.lifetime  and self.lifetime > 0)
        
        
    @property
    def is_authenticated(self) -> bool:
        user_model = self.get_user()
        return bool (user_model and user_model.id and self.has_lifetime)


    @property
    def is_manager(self) -> bool:
        user_model = self.get_user()
        return bool (user_model and user_model.role and user_model.role.is_admin and self.has_lifetime)
    
    
    @property
    def is_admin(self) -> bool:
        user_model = self.get_user()
        return bool (user_model and user_model.is_admin and self.lifetime  and self.lifetime > 0)


from typing import Optional
from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.bags.msg_bag import MsgBag
from nkunyim_util.models.application_model import ApplicationModel
from nkunyim_util.caches.application_cache import ApplicationCache
from nkunyim_util.api.nkunyim_api_client import NkunyimApiClient
from nkunyim_util.models.msg_model import MsgFactory, MsgLevel


from .signals import app_data_updated


class ApplicationService:
    
    def __init__(self, req: HttpRequest, session_key: str) -> None:
        self.cache = ApplicationCache(key=f"app.{session_key}")
        self.req = req
        self.msgs = MsgBag()


    def _get_from_api(self) -> Optional[dict]:
        try:
            modules = []
            domain = '.'.join(self.req.get_host().rsplit('.', 2)[-2:]).lower()
            client = NkunyimApiClient(req=self.req, name=settings.MARKET_SERVICE)
            response = client.get(path=f"/api/applications/session/?domain={domain}")
            if not response.ok:
                self.msgs.add(MsgFactory.create(level=MsgLevel.WARNING, message="Failed to retrieve application data from API."))
                return None
            
            json_data = response.json()
            application = json_data['application']
            packets = json_data.get('packets', None)
            if packets:
                for packet in packets:
                    modules.append({
                        "id": packet['module']['id'],
                        "name": packet['module']['name'],
                        "title": packet['module']['title'],
                        "version": packet['version']['code'],
                    })

            return {
                **application,
                "modules": modules,
            }
        except:
            return None


    def _make(self) -> ApplicationModel:
        model_data = self._get_from_api()
        if not model_data:
            model_data = dict(settings.NKUNYIM_DEFAULT_APP)
    
        app_model = ApplicationModel(**model_data) # type: ignore
        self.cache.set(model=app_model, timeout=60 * 60 * 24)
        
        # Inform interested parties
        app_data_updated.send(sender=ApplicationModel, instance=app_model)
        
        return app_model
    
    
    def get(self) -> ApplicationModel:
        return self.cache.get() or self._make()
    